package media

import (
	"bytes"
	"encoding/binary"
	"fmt"
	"io"
)

type FetchAudioProgress struct {
	percentComplete float32
	Peaks           []int16
}

type FetchAudioProgressReader interface {
	Read() (FetchAudioProgress, error)
	Close() error
}

// fetchAudioProgressReader accepts a byte stream containing little endian
// signed int16s and, given a target number of bins, emits a stream of peaks
// corresponding to each channel of the audio data.
type fetchAudioProgressReader struct {
	channels     int
	framesPerBin int

	samples   []int16
	currPeaks []int16
	currCount int
	total     int
	progress  chan FetchAudioProgress
	errorChan chan error
}

// TODO: validate inputs, debugging is confusing otherwise
func newFetchAudioProgressReader(expFrames int64, channels, numBins int) *fetchAudioProgressReader {
	return &fetchAudioProgressReader{
		channels:     channels,
		framesPerBin: int(expFrames / int64(numBins)),
		samples:      make([]int16, 8_192),
		currPeaks:    make([]int16, channels),
		progress:     make(chan FetchAudioProgress),
		errorChan:    make(chan error, 1),
	}
}

func (w *fetchAudioProgressReader) Abort(err error) {
	w.errorChan <- err
}

func (w *fetchAudioProgressReader) Close() error {
	close(w.progress)
	return nil
}

func (w *fetchAudioProgressReader) Write(p []byte) (int, error) {
	// expand our target slice if it is of insufficient size:
	numSamples := len(p) / SizeOfInt16
	if len(w.samples) < numSamples {
		w.samples = append(w.samples, make([]int16, numSamples-len(w.samples))...)
	}

	samples := w.samples[:numSamples]

	if err := binary.Read(bytes.NewReader(p), binary.LittleEndian, samples); err != nil {
		return 0, fmt.Errorf("error parsing samples: %v", err)
	}

	for i := 0; i < len(samples); i += w.channels {
		for j := 0; j < w.channels; j++ {
			samp := samples[i+j]
			if samp < 0 {
				samp = -samp
			}
			if samp > w.currPeaks[j] {
				w.currPeaks[j] = samp
			}
		}
		w.currCount++
		if w.currCount == w.framesPerBin {
			w.nextBin()
		}
	}

	return len(p), nil
}

func (w *fetchAudioProgressReader) nextBin() {
	var progress FetchAudioProgress
	// TODO: avoid an allocation?
	progress.Peaks = append(progress.Peaks, w.currPeaks...)

	w.progress <- progress

	w.currCount = 0
	// log.Printf("got peak for %d frames, which is equal to target of %d frames per bin, %d total bins processed, peaks: %+v", w.currCount, w.framesPerBin, w.total+1, w.currPeaks)
	for i := 0; i < len(w.currPeaks); i++ {
		w.currPeaks[i] = 0
	}
	w.total++
}

func (w *fetchAudioProgressReader) Read() (FetchAudioProgress, error) {
	for {
		select {
		case progress, ok := <-w.progress:
			if !ok {
				return FetchAudioProgress{}, io.EOF
			}
			return FetchAudioProgress{Peaks: progress.Peaks}, nil
		case err := <-w.errorChan:
			return FetchAudioProgress{}, fmt.Errorf("error waiting for progress: %v", err)
		}
	}
}
